/***********************************************************************************

	Copyright (C) 2009 Ahmet Öztürk (aoz_2@yahoo.com)

	This file is part of Lifeograph.

	Lifeograph is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Lifeograph is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

***********************************************************************************/


#ifndef LIFEOGRAPH_TEXTVIEW_HEADER
#define LIFEOGRAPH_TEXTVIEW_HEADER


#include <deque>

extern "C"
{
#include <enchant/enchant.h>
}

#include "helpers.hpp"
#include "undo.hpp"
#include "diarydata.hpp"
#include "entry.hpp"
#include "entry_parser.hpp"


namespace LIFEO
{

const int		UNDO_MERGE_TIMEOUT = 3;

enum LinkStatus
{
	LS_OK,
	LS_ENTRY_UNAVAILABLE,
	LS_INVALID,		// separator: to check a valid entry link: linkstatus < LS_INVALID
	LS_CYCLIC,

	LS_FILE_OK,
	LS_FILE_INVALID,
	LS_FILE_UNAVAILABLE,
	LS_FILE_UNKNOWN
};

// UNDO ============================================================================================
class UndoEdit : public Undoable
{
	public:
							UndoEdit( int position,
                                      const Glib::ustring& text,
                                      UndoableType type )
		: Undoable( "placeholder", type ), m_position( position ), m_text( text ) { }
		virtual             ~UndoEdit() {}   //needed because of virtual methods

		unsigned int		get_position (void) const
		{
			return m_position;
		}

		Glib::ustring		get_text( void ) const
		{
			return m_text;
		}

	protected:
		virtual void		merge( Undoable* ) = 0;

		virtual void		undo( void ) = 0;
		virtual void		redo( void ) = 0;

		void				erase( void )
		{
			Gtk::TextIter iter_start = m_ptr2buffer->get_iter_at_offset( m_position );
			Gtk::TextIter iter_end = m_ptr2buffer->get_iter_at_offset( m_position +
													m_text.size() );
			m_ptr2buffer->erase( iter_start, iter_end );
		}

		void				insert( void )
		{
			Gtk::TextIter iterator = m_ptr2buffer->get_iter_at_offset( m_position );
			m_ptr2buffer->insert( iterator, m_text );
		}

		unsigned int		m_position;
		Glib::ustring		m_text;

	private:
		static Gtk::TextBuffer		*m_ptr2buffer;

	friend class TextbufferDiary;
};

class UndoInsert :public UndoEdit
{
	public:
							UndoInsert( int position, const Glib::ustring& text )
		: UndoEdit( position, text, UT_INSERT_TEXT ) { }

		bool				can_merge( const Undoable* action ) const
		{
			if( ( action->get_time_sec() - m_time.tv_sec ) < UNDO_MERGE_TIMEOUT )
				if( action->get_type() == m_type )
					if( dynamic_cast< const UndoInsert* >( action )->get_position()
							== m_position + m_text.size() )
					return true;
			return false;
		}

		void				merge (Undoable* action)
		{
			m_text += dynamic_cast< UndoInsert* >( action )->get_text();
		}

	protected:
		void				undo( void )
		{
			erase();
		}
		void				redo( void )
		{
			insert();
		}
};

class UndoErase : public UndoEdit
{
	public:
							UndoErase( int position, const Glib::ustring& text )
		: UndoEdit( position, text, UT_ERASE_TEXT ) { }

		bool				can_merge (const Undoable* action) const
		{
			if ( ( action->get_time_sec() - m_time.tv_sec ) < UNDO_MERGE_TIMEOUT )
				if ( action->get_type() == m_type )
				{
					const UndoErase * action_erase =
						dynamic_cast< const UndoErase* >( action );
					if ( action_erase->get_position() +
							action_erase->get_text().size()
								== m_position )
						return true;
				}
			return false;
		}

		void				merge (Undoable* action)
		{
			UndoErase * action_erase = dynamic_cast< UndoErase* >( action );
			m_text.insert( 0, action_erase->get_text() );
			m_position = action_erase->get_position();
		}

	protected:
		void				undo( void )
		{
			insert();
		}
		void				redo( void )
		{
			erase();
		}

};

// LINKS ===========================================================================================
class Link
{
	public:
		virtual						~Link();
		virtual void				go( void ) = 0;

		// position of link in the buffer:
		Glib::RefPtr< Gtk::TextMark >
									m_mark_start;
		Glib::RefPtr< Gtk::TextMark >
									m_mark_end;

	protected:
									Link(	const Glib::RefPtr< Gtk::TextMark >&,
											const Glib::RefPtr< Gtk::TextMark >& );
};

class LinkID : public Link
{
	public:
		typedef sigc::signal< void, Date >
									Signal_void_Date;

									LinkID( const Glib::RefPtr< Gtk::TextMark >&,
											const Glib::RefPtr< Gtk::TextMark >&,
											DEID );
		void						go( void );

		static Signal_void_Date		signal_activated( void )
		{ return m_signal_activated; }

		DEID						m_id;

	protected:
		static Signal_void_Date		m_signal_activated;
};

class LinkEntry : public Link
{
	public:
									LinkEntry(	const Glib::RefPtr< Gtk::TextMark >&,
												const Glib::RefPtr< Gtk::TextMark >&,
												Date );
		void						go( void );

		Date						m_date;

	protected:
        static Gtk::Menu            *menu_link;
};

class LinkUri : public Link
{
	public:
									LinkUri(	const Glib::RefPtr< Gtk::TextMark >&,
												const Glib::RefPtr< Gtk::TextMark >&,
												const std::string& );
		void						go( void );

		std::string					m_url;
		static Gtk::TextBuffer		*m_ptr2buffer;
};

class LinkCheck : public Link
{
	public:
									LinkCheck( const Glib::RefPtr< Gtk::TextMark >&,
											   const Glib::RefPtr< Gtk::TextMark >&,
											   unsigned int = 0 );
		void						go( void );

		const static Glib::ustring	boxes;
		unsigned int				m_state_index;
};

class InlineImage
{
    public:
        InlineImage( const Glib::RefPtr< Gtk::TextMark >& );
        ~InlineImage();

        Glib::RefPtr< Gtk::TextMark > m_mark;
};

// NEW LINE PREDICATE ==============================================================================
class PredicateNL
{
  public:
    bool operator()( gunichar ch ) const
    {
        return( ch == '\n' );
    }
};

// TEXTBUFFER ======================================================================================
class TextbufferDiary : public Gtk::TextBuffer, public EntryParser
{
	public:
		typedef void ( TextbufferDiary::*ParsingApplierFn )( void );

//		typedef sigc::signal< LinkStatus, Date >
//									Signal_LinkStatus_Date;
		typedef sigc::signal< void, const Glib::ustring& >
									Signal_void_ustring;

									TextbufferDiary( void );
		void						set_textview( Gtk::TextView *ptr2textview )
		{ m_ptr2textview = ptr2textview; }
		void						handle_logout( void );
		void						handle_login( void );

		Link*						get_link( int ) const;
		Link*						get_link( const Gtk::TextIter& ) const;

		void						set_richtext( Entry* );

		void						set_search_str( const Glib::ustring& );
		bool						select_searchstr_previous( void );
		bool						select_searchstr_next( void );

		void						toggle_format(	Glib::RefPtr< Tag >,
													const Glib::ustring& );
		void						toggle_bold( void );
		void						toggle_italic( void );
		void						toggle_highlight( void );
		void						toggle_strikethrough( void );

        void                        set_comment_visibility( bool );
        bool                        get_comment_visibility( void ) const
        { return m_tag_comment->property_invisible(); }

        static void                 calculate_para_bounds( Gtk::TextIter&, Gtk::TextIter& );
        bool                        calculate_multiline_selection_bounds( Gtk::TextIter&,
                                                                          Gtk::TextIter& );
		void						handle_indent( void );
		void						handle_unindent( void );

		void						add_bullet( void );
		void						add_checkbox( void );
		void						add_empty_line_above( void );
		void						remove_empty_line_above( void );
        void                        move_line_up( void );
        void                        move_line_down( void );
		void						insert_link( DiaryElement* );
		void						insert_time_stamp( void );

        std::string                 set_language( std::string );

		void						set_theme( const Theme* );

		void						handle_menu( Gtk::Menu* );

	protected:
		void						on_insert( const Gtk::TextIter&,
											   const Glib::ustring&,
											   int );
		void						on_erase( const Gtk::TextIter&,
											  const Gtk::TextIter& );

		/*void						on_apply_tag( const Glib::RefPtr< TextBuffer::Tag >&,
												  const Gtk::TextIter&,
												  const Gtk::TextIter& );*/

		void						on_mark_set( const TextBuffer::iterator&,
												 const Glib::RefPtr< TextBuffer::Mark >& );

		void						process_space( void );
		bool						process_new_line( void );

		// PARSING
        void                        parse( const Gtk::TextIter&, const Gtk::TextIter& );
        void                        reparse( void );

		void						apply_hidden_link_tags( Gtk::TextIter&,
		                                    const Glib::RefPtr< Tag >& );

		void						apply_heading( void );
		void						apply_subheading( void );
		void						apply_bold( void );
		void						apply_italic( void );
		void						apply_strikethrough( void );
		void						apply_highlight( void );
        void                        apply_comment( void );
        void                        apply_ignore( void );
		void						apply_link( void );
		void						apply_link_date( void );
		void						apply_link_id( void );
		void						apply_check_ccl( void );
		void						apply_check_unf( void );
		void						apply_check_fin( void );
        void                        apply_match( void );
		void                        apply_markup( const Glib::RefPtr<Tag>& );

		gunichar                    get_char_at( int );

        void                        handle_word( void );

        void                        handle_spellcheck_toggled( void );
        void                        handle_language_changed( const std::string&,
                                                             const Gtk::RadioMenuItem* );
        Gtk::Menu*                  build_languages_menu( void );
        void                        add_suggestion_menus( const Glib::ustring&, Gtk::Menu* );
        void                        ignore_misspelled_word( const Glib::ustring& );
        void                        replace_misspelled_word( const Glib::ustring& );
        void                        add_word_to_dictionary( const Glib::ustring& );

		Glib::ustring::size_type	parser_open_tag_begin;
		Glib::ustring::size_type	parser_open_tag_end;
		bool						parser_open_tag_is_link;
		Glib::ustring::size_type	parser_offset_insert_prev;

        static PredicateNL          s_predicate_nl;

		// LINKS
		void						clear_links( int, int );
		void						clear_links( void );

		void                        clear_images( int, int& );
        void                        clear_images( void );

        // TAGS
        // TODO: move these program-wide objects out of this class as...
        // ...there may be other instances of TextbufferDiary in the future
        Glib::RefPtr< Tag >         m_tag_heading;
        Glib::RefPtr< Tag >         m_tag_subheading;
        Glib::RefPtr< Tag >         m_tag_match;
        Glib::RefPtr< Tag >         m_tag_markup;
        Glib::RefPtr< Tag >         m_tag_markup_link;
        Glib::RefPtr< Tag >         m_tag_hidden;
        Glib::RefPtr< Tag >         m_tag_bold;
        Glib::RefPtr< Tag >         m_tag_italic;
        static Glib::RefPtr< Tag >  m_tag_strikethrough;
        static Glib::RefPtr< Tag >  m_tag_highlight;
        Glib::RefPtr< Tag >         m_tag_comment;
        Glib::RefPtr< Tag >         m_tag_region;
        Glib::RefPtr< Tag >         m_tag_link;
        Glib::RefPtr< Tag >         m_tag_link_broken;
        Glib::RefPtr< Tag >         m_tag_link_hidden;
        static Glib::RefPtr< Tag >  m_tag_checkbox;
        Glib::RefPtr< Tag >         m_tag_image;
        Glib::RefPtr< Tag >         m_tag_misspelled;

        typedef std::vector< Glib::RefPtr< Tag > > VecTags;
        VecTags                     m_tags;

		// THEMING
		static Gdk::RGBA			m_theme_color_match;
		static Gdk::RGBA			m_theme_color_markup;
		static Gdk::RGBA			m_theme_color_link;
		static Gdk::RGBA			m_theme_color_linkbroken;

		// OTHER VARIABLES
		bool						m_flag_settextoperation;
        int                         m_flag_ongoing_operation;   // Incremental to allow nesting
		bool						m_flag_parsing;
		Entry						*m_ptr2entry;
		Gtk::TextView				*m_ptr2textview;

		typedef std::list< Link* >	ListLinks;
		ListLinks					m_list_links;
		typedef std::set< InlineImage* >
		                            ListImages;
		ListImages                  m_list_images;

		// ENCHANT SPELLCHECKER
        EnchantDict                 *m_enchant_dict;
        int                         m_spell_suggest_offset;

	friend class DialogTheme;
	friend class EntryView;
	friend class LinkCheck;
	friend class TextviewDiary;
};

// TEXTVIEW ========================================================================================
class TextviewDiary : public Gtk::TextView
{
	public:
									TextviewDiary( BaseObjectType*,
												   const Glib::RefPtr<Gtk::Builder>& );

        void                        set_richtext( Entry* );

        TextbufferDiary             *m_buffer;

	protected:
		void						update_link( void );
		virtual bool				on_motion_notify_event( GdkEventMotion* );
		virtual bool				on_button_press_event( GdkEventButton* );
		virtual bool				on_button_release_event( GdkEventButton* );
		virtual bool				on_key_press_event( GdkEventKey* );
		virtual bool				on_key_release_event( GdkEventKey* );
		//~ virtual void				on_populate_popup( Gtk::Menu* );
//		virtual void				on_style_changed( const Glib::RefPtr< Gtk::Style >& );

	private:
        const Glib::RefPtr< Gdk::Cursor >
                                    m_cursor_hand;
        const Glib::RefPtr< Gdk::Cursor >
                                    m_cursor_xterm;
        const Glib::RefPtr< Gdk::Cursor >
                                    *m_ptr2cursor_last;
		Link						*m_link_hovered;
};

} // end of namespace LIFEO

#endif
